#include <stdlib.h>
#include <time.h>	//fr time in srand
#include "particlesys.h"

////////////////////////////////////////////////////////////////////////////////
// CParticelSys - Funktionen
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////
// Konstruktor und Destruktor //
////////////////////////////////
CParticelSys::CParticelSys(int feldbreite,int feldhoehe,short schweifl,short richtungen,short farbtiefe, bool fullscreen)
{
	int i,j;

	// Zufallsgenerator starten
	srand((unsigned)time( NULL ));

	Uint32 flag=SDL_SWSURFACE|SDL_DOUBLEBUF;
	if(fullscreen) flag|=SDL_FULLSCREEN;
	else farbtiefe=0;

	// Fenster erzeugen
	pS = SDL_SetVideoMode(feldbreite,feldhoehe,farbtiefe,flag);
	if(!pS)
	{
		fprintf(stderr,"Konnte Bildschirmmodus nicht setzten: %s\n",SDL_GetError());
		exit(1);
	}

	// Speicher fr Farbtabelle reservieren
	CParticel::farben = new Uint32*[7];
	for(i=0;i<7;i++)
		CParticel::farben[i] = new Uint32[schweifl];

	if (CParticel::farben == 0)
	{
		fprintf(stderr,"Konnte Speicher fr die Farbttabelle nicht reservieren!");
		exit(2);
	}

	// Farben fr die Particel	
	Uint8 xt=255/(schweifl-1);
	for (i=0;i<schweifl;i++)
		for(j=0;j<7;j++)
			switch (j)
			{
				case 0: CParticel::farben[j][i]=SDL_MapRGB(pS->format,i*xt,i*xt,i*xt);break;// Schwarz
				case 1: CParticel::farben[j][i]=SDL_MapRGB(pS->format, 255,i*xt,i*xt);break;// Rot
				case 2: CParticel::farben[j][i]=SDL_MapRGB(pS->format,i*xt, 255,i*xt);break;// Grn
				case 3: CParticel::farben[j][i]=SDL_MapRGB(pS->format,i*xt,i*xt, 255);break;// Blau
				case 4: CParticel::farben[j][i]=SDL_MapRGB(pS->format, 255, 255,i*xt);break;// Gelb	
				case 5: CParticel::farben[j][i]=SDL_MapRGB(pS->format, 255,i*xt, 255);break;// Magenta
				case 6: CParticel::farben[j][i]=SDL_MapRGB(pS->format,i*xt, 255, 255);break;// Cyan
			}

	// Feld fr die Particel-Kontrolle erzeugen
	feld = new Uint32*[feldbreite];
	for(i=0;i<feldbreite;i++)
		feld[i]= new Uint32[feldhoehe];

	if (feld == 0)
	{
		fprintf(stderr,"Konnte speicher fr das Feld nicht reservieren!");
		exit(2);
	}

	// Feld in Particel lschen
	for(i=0;i<feldbreite;i++)
		for(j=0;j<feldhoehe;j++)
			feld[i][j] = 0xffffffff;

	// Feld an Particel bergeben
	CParticel::feld=feld;
	
	// Wichtige statiche werte fr die Particel
	CParticel::feldbreite=this->feldbreite=feldbreite;
	CParticel::feldhoehe=this->feldhoehe=feldhoehe;
	CParticel::schweifl=schweifl;
	CParticel::richtungen=richtungen;
}

CParticelSys::~CParticelSys()
{
	short i;

	// alle Particel aus dem Speicher lschen
	alle_par_weg();

	// speicher von feld und farben wieder frei geben
	for (i=0;i<feldbreite;i++)
		delete [] feld[i];
	delete [] feld;

	for(i=0;i<7;i++)
		delete [] CParticel::farben[i];
	delete [] CParticel::farben;

}

//////////////////////////
// Zeichen der Particel //
//////////////////////////
void CParticelSys::zeichnen()
{
	static const Uint32 farbe_weis = SDL_MapRGB(pS->format,255,255,255);
	short i,j;
	
	if (par.size()>0)
	{
		// Feld in Particel lschen
		for(i=0;i<feldbreite;i++)
			for(j=0;j<feldhoehe;j++)
				feld[i][j] = 0xffffffff;

		// Logik des partikels ausfhren
		for (short i=0;i<par.size();i++)
				par[i]->tick();
		
		// Hintergrund wei zeichnen
		if(SDL_FillRect(pS,0,farbe_weis))
		{
			fprintf(stderr,"Fehler bei FillRect: %s\n",SDL_GetError());
			exit(6);
		}
		
		// alle Partikel Zeichnen
		if(!SDL_LockSurface(pS))
		{
			// Hintergrund wei
			

			// Particel zeichnen
			for(i=0;i<feldbreite;i++)
				for(j=0;j<feldhoehe;j++)
					if(feld[i][j]!=0xffffffff) SetPixel(i,j);
			
			// Surface wieder freigeben
			SDL_UnlockSurface(pS);
		}
		else 
		{
			fprintf(stderr,"SDL_LockSurface fehlgeschlagen.\n");
			exit(3);
		}
	}
	else
	{
		SDL_FillRect(pS,0,farbe_weis);
	}

	// auf Bildschirm bertragen
	SDL_Flip(pS);	
}

void CParticelSys::SetPixel(short x, short y)
{
	if(x<0||x>pS->w||y<0||y>pS->h) return;

	Uint32 color = feld[x][y];

	int bpp=pS->format->BytesPerPixel;
	Uint8 *p=(Uint8 *)pS->pixels + y*pS->pitch + x*bpp;

	switch(bpp)
	{
	case 1: *p = color;break;
	case 2: *(Uint16 *)p = color;break;
	case 4: *(Uint32 *)p = color; break;
	case 3:
		#if SDL_BYTEORDER==SDL_BIG_ENDIAN
		p[0]=(color>>16)&0xff;
		p[1]=(color>>8)&0xff;
		p[2]=color&0xff;
		#else
		p[0]=color&0xff;
		p[1]=(color>>8)&0xff;
		p[2]=(color>>16)&0xff;
		#endif
		break;
	}
}

/////////////////////////////////////////////////////
// Funktionen fr Hinzufgen/Lschen von Particeln //
/////////////////////////////////////////////////////
void CParticelSys::neue_particel(short x,short y,int anzahl, short farbe)
{
	for (;anzahl>0;anzahl--)
		par.push_back(new CParticel(x,y,farbe));
}

void CParticelSys::alle_par_weg()
{
	for (int i=par.size();i>0;i=par.size())
	{
		delete par[i-1];// lscht das Objekt
		par.pop_back();	// lscht den Pointer(des gelschten Objektes) aus der Liste
	}
}



////////////////////////////////////////////////////////////////////////////////////////////
// CParticel - Funktionen
////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////
// Grundinitialisierung der Statichen variablen //
//////////////////////////////////////////////////	
Uint32** CParticel::feld=0;
Uint32** CParticel::farben=0;
short CParticel::feldbreite=100;
short CParticel::feldhoehe=100;
short CParticel::richtungen=1;
short CParticel::schweifl=2;
bool CParticel::const_geschw=false;
bool CParticel::zuf_richtwechsel=false;

////////////////////////////////
// Konstruktor und Destruktor //
////////////////////////////////
CParticel::CParticel(short x, short y, short farbe)
{
	// Speicher fr Schweif reservieren
	schweif = 0;
	schweif = new strSchweif[schweifl];
	if (schweif == 0)
	{
		fprintf(stderr,"Konnte Speicher fr den Schweif nicht reservieren!");
		exit(2);
	}

	// x und y auf richtigkeit berprfen, anpassen
	if(x<0) x=0;
	if(x>=feldbreite) x=feldbreite-1;
	if(y<0) y=0;
	if(y>=feldhoehe) y=feldhoehe-1;

	
	// Startposition setzen
	schweif[0].x=x;
	schweif[0].y=y;

	for (short i=0;i<schweifl;i++)
		schweif[i]=schweif[0];

	// Bewegungsrichtung setzten
	zufall(dx,dy);
	
	// Farbe setzen
	this->farbe = farbe;

	// zufllige richtungsnderung
	count=10+rand()%100;
}

CParticel::~CParticel()
{
	// Speicher fr Schweif freigeben
	delete [] schweif;
}

/////////////////////////////////////////////////////////////////
// Funktion fr die Fortbewegung und neuzeichnen des Particels //
/////////////////////////////////////////////////////////////////
void CParticel::tick_alle(short &dx, short &dy,short &count, short &farbe, strSchweif* schweif)
{
	// Zufllige Richtungsnderung
	if (zuf_richtwechsel)
		if (--count==0)
		{
			count=rand()%100;
			zufall(dx,dy);
		}

	// Position des Particel
	short x=schweif[0].x;
	short y=schweif[0].y;

	// Prfen ob neue Position zulssig
	if (((x+dx) >= feldbreite)||((x+dx) <0))
	{
		if (!const_geschw) zufall(dx,dy);
		else dx=-dx;
	}
	else x+=dx;

	if (((y+dy) >= feldhoehe)||((y+dy) <0))
	{
		if (!const_geschw) zufall(dx,dy);
		else dy=-dy;
	}
	else y+=dy;

	// Schweif nachrcken, zeichnen
	for (short i=schweifl-1;i>0;--i)
	{
		schweif[i]= schweif[i-1];
		if (feld[schweif[i].x][schweif[i].y]==0xffffffff)
			feld[schweif[i].x][schweif[i].y]=farben[farbe][i];
	}

	// neue Position setzen zeichnen
	schweif[0].x=x;
	schweif[0].y=y;	
	if (feld[schweif[i].x][schweif[i].y]==0xffffffff)
		feld[schweif[i].x][schweif[i].y]=farben[farbe][i];

}

void CParticel::tick()
{
	tick_alle(this->dx,this->dy,this->count,this->farbe,this->schweif);
}

/*void CParticel::tick()
{
	// Zufllige Richtungsnderung
	if (zuf_richtwechsel)
		if (--count==0)
		{
			count=rand()%100;
			zufall(dx,dy);
		}

	// Position des Particel
	short x=schweif[0].x;
	short y=schweif[0].y;

	// Prfen ob neue Position zulssig
	if (((x+dx) >= feldbreite)||((x+dx) <0))
	{
		if (!const_geschw) zufall(dx,dy);
		else dx=-dx;
	}
	else x+=dx;

	if (((y+dy) >= feldhoehe)||((y+dy) <0))
	{
		if (!const_geschw) zufall(dx,dy);
		else dy=-dy;
	}
	else y+=dy;

	// Schweif nachrcken, zeichnen
	for (short i=schweifl-1;i>0;--i)
	{
		schweif[i]= schweif[i-1];
		if (feld[schweif[i].x][schweif[i].y]==0xffffffff)
			feld[schweif[i].x][schweif[i].y]=farben[farbe][i];
	}

	// neue Position setzen zeichnen
	schweif[0].x=x;
	schweif[0].y=y;	
	if (feld[schweif[i].x][schweif[i].y]==0xffffffff)
		feld[schweif[i].x][schweif[i].y]=farben[farbe][i];
}
*/
//////////////////////////////////////////////////////////
// Statiche Funktionen, auf die alle Particel zugreifen //
//////////////////////////////////////////////////////////
void CParticel::zufall(short &dx, short &dy)
{
	// neue werte setzen
	dx=richtungen-(rand()%(richtungen*2+1));
	dy=richtungen-(rand()%(richtungen*2+1));
	// prfen ob sich der particel auch bewegt (dx==dy==0 -> keine bewegung)
	if ((dx==0)&&(dy==0)) zufall(dx,dy);
}
